/*

  File LUmat.cpp:

  contains the definition of the member functions for the LUmat class.
  There is a function to define a generic nxn matrix, to factorize LU 
  the matrix, to solve a system with a given RHS, and work functions
  to set/get the elements of the matrix.

*/


#include "LUmat.h"


/*

 LUmat(): class constructor

*/

LUmat::LUmat( int N )
{
  // The class variable dim is initialized to N
  dim = N;

  // Define the original and factorized matrices as vectors of pointers
  mat = new double * [N];
  for ( int i = 0; i < N; i++ )
    {
      mat[i] = new double[N];
    }

  // Defines the temporary vect with the same dimension N
  tmp = new double [N];
}


/*

  luFact( ): LU factorization of mat[dim][dim] with the Crout algorithm.
  Please, note that the matrix mat is overwritten during the factorization
  in order to save memory!

*/


void LUmat::luFact( )
{

  double sum;

  // First row
  // u_1j = a_1j / l_11, for j = 2, ..., n

  for ( int j = 1; j < dim ; j++ ) mat[0][j] = mat[0][j] / mat[0][0];

  // Rows for i = 2,...,n

  for ( int i = 1; i < dim; i++ )
    {

      // j = 2, ..., i
      for ( int j = 1; j <= i; j++ )
	{
	  sum = 0.0;
	  // Computes: Sum_k=1^k=j-1 l_ik * u_kj
	  for ( int k = 0; k < j; k++ ) sum += mat[i][k] * mat[k][j];
	  // l_ij = a_ij - Sum...
	  mat[i][j] = mat[i][j] - sum;
	}

      // j = i+1, ..., n
      for ( int j = i + 1; j < dim; j++ )
	{
	  sum = 0.0;
	  // Computes: Sum_k=1^k=i-1 l_ik * u_kj
	  for ( int k = 0; k < i; k++ ) sum += mat[i][k] * mat[k][j];
	  // u_ij = ( a_ij - Sum ... ) / l_ii
	  mat[i][j] = ( mat[i][j] - sum ) / mat[i][i];
	}

    }

}


/*

  luSolve( x, b ): inverts the system mat*x=b by using the FS and BS

*/


void LUmat::luSolve( LUvect x, LUvect b )
{

  double sum;

  // Forward substitution
  // Solution of: L*y = b
  // We use the tmp vector for storing the result of the FS (y)! 

  // First element: y_1 = b_1 / l_11

  tmp[0] = b.getElement( 1 ) / mat[0][0];

  // i = 2, ..., n
  
  for ( int i = 2; i <= dim; i++ )
    {
      sum = 0.0;
      // Computes: Sum_j=1^j=i-1 l_ij * y_j
      for ( int j = 1; j <= i - 1; j++ ) sum += mat[i-1][j-1] * tmp[j-1];
      // y_i = ( b_i - Sum ... ) / l_ii
      tmp[i-1] = ( b.getElement( i ) - sum ) / mat[i-1][i-1];
    }

  // Backward substitution
  // Solution of: U*x = y
  
  // First element: x_n = y_n / u_nn

  x.setElement( dim, tmp[dim-1] );

  // i = n -1, ..., 1

  for ( int i = dim - 1; i >= 1; i-- )
    {
      sum = 0.0;
      // Computes: Sum_j=i+1^j=n u_ij * x_j
      for ( int j = i + 1; j <= dim; j++ ) sum += mat[i-1][j-1] * x.getElement( j );
      // x_i = ( b_i - Sum ... ) / u_ii
      x.setElement( i, tmp[i-1] - sum );
    }

}


/*

  setElement: set the ij-th element of the matrix to the given value

*/

void LUmat::setElement( int i, int j, double a_ij )
{
  if ( ( ( i > 0 ) && ( i <= dim ) ) && ( ( j > 0 ) && ( j <= dim ) ) )
    {
      mat[i-1][j-1] = a_ij;
    }
}


/*

  getElement: return the value of ij-th element of the matrix

*/

double LUmat::getElement( int i, int j )
{
  if ( ( ( i > 0 ) && ( i <= dim ) ) && ( ( j > 0 ) && ( j <= dim ) ) )
    return( mat[i-1][j-1] );
}


/*

  matVectProduct: computes the product of the current matrix by a vector u 
  and returns the result in another vector v.

*/

void LUmat::matVectProduct( LUvect u, LUvect v )
{
  double sum;

  for ( int i = 1; i <= dim; i++ )
    {
      sum = 0.0;
      for ( int j = 1; j <= dim; j++ )
	{
	  sum += mat[i-1][j-1] * u.getElement( j );
	}
      v.setElement( i, sum );
    }
}



/*

  matMatProduct: computes the product of the current matrix by another matrix B
  and returns the result in a third matrix C

*/

void LUmat::matMatProduct( LUmat B, LUmat C )
{
  double sum;

  for ( int i = 1; i <= dim; i++ )
    {
      for ( int j = 1; j <= dim; j++ )
	{
	  sum = 0.0;
	  for ( int k = 1; k <= dim; k++ )
	    {
	      sum += mat[i-1][k-1] * B.mat[k-1][j-1];
	    }
	  C.mat[i-1][j-1] = sum;
	}
    }
}



/*

  duplicateMat: copies the elements of a matrix in the current matrix

*/

void LUmat::duplicateMat( LUmat B )
{
  for ( int i = 1; i <= dim; i++ )
    for ( int j = 1; j <= dim; j++ )
      this->mat[i-1][j-1] = B.mat[i-1][j-1];
}
