/*

  File Sys_scal.ccp:

  To verify that the complexity of the LU algorithm is ~ O(n^3)
  we solve a generic NxN system by defining a matrix of the 
  coefficients as:
  a_ij = 1                 if i  = j;
  a_ij = 1.0/(1.0+i+j)     if i /= j

  Note that THIS IS a diagonally dominant matrix!

  Then, we change N and verify that the number of operations changes as N^3.

  Please, comment out all prints and verifications to estimate the time!

*/


#include <iostream>
using namespace std;

#include "LUvect.h"
#include "LUmat.h"


int main()
{
  cout << "Solve a system of linear equations through the LU factorization" << endl;

  const int N = 100;


  // Instantiates the objects omat and mat of class LUmat and define them

  LUmat omyMat = LUmat( N );
  LUmat myMat = LUmat( N );
  for ( int i = 1; i <= N; i++ )
    for ( int j = 1; j <= N; j++ )
      {
        if ( i == j ) omyMat.setElement( i, j, 1.0);
        else
          omyMat.setElement( i, j, 1.0 / ( 1.0 + i + j ) );
      }
  myMat.duplicateMat( omyMat );
  

  // Prints the matrix of coefficients on the screen

  cout.precision( 15 );     // Sets output precision with 15 decimal figures
  cout << "Initial matrix:" << endl;
  cout << "--------------------------------------------------------------------------" << endl;
  for ( int i = 1; i <= N; i++ )
    {
      for ( int j = 1; j <= N; j++ ) cout << "   " << omyMat.getElement( i, j );
      cout << endl;
    }
  cout << "--------------------------------------------------------------------------" << endl;


  // Lu factorization of the matrix

  myMat.luFact( );

  
  // Defines the vectors for the solution (x) and known terms (b)

  LUvect x = LUvect( N );
  LUvect b = LUvect( N );

  for ( int i = 1; i <= N; i++ )
    {
      b.setElement( i, ( i -1 ) * 1.0 );
    }


  // Prints on the screen the RHS of the system

  cout << "RHS vector:" << endl;
  cout << "--------------------------------------------------------------------------" << endl;
  for ( int i = 1; i <= N; i++ ) cout << "   " << b.getElement( i );
  cout << endl;
  cout << "--------------------------------------------------------------------------" << endl;


  // Solves the system

  myMat.luSolve( x, b );

  
  // Prints on the screen the solution of the system

  cout << "Vector of the solution:" << endl;
  cout << "--------------------------------------------------------------------------" << endl;
  for ( int i = 1; i <= N; i++ ) cout << "   " << x.getElement( i );
  cout << endl;
  cout << "--------------------------------------------------------------------------" << endl;


  // Verify the result. The product A*x is stored in b

  cout << "Product of the matrix by the solution vector:" << endl;
  omyMat.matVectProduct( x, b );
  cout << "--------------------------------------------------------------------------" << endl;
  for ( int i = 1; i <= N; i++ ) cout << "   " << b.getElement( i );
  cout << endl;
  cout << "--------------------------------------------------------------------------" << endl;

  return 0;
}
